# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_fact.ipynb.

# %% auto 0
__all__ = ['SI_FACTORS', 'fmeta', 'fact', 'zfact', 'pfact', 'nfact', 'one', 'deca', 'hecto', 'kilo', 'mega', 'giga', 'tera',
           'peta', 'exa', 'zetta', 'yotta', 'ronna', 'quetta', 'deci', 'centi', 'milli', 'micro', 'nano', 'pico',
           'femto', 'atto', 'zepto', 'yocto', 'ronto', 'quecto']

# %% ../nbs/04_fact.ipynb 3
from typing import Any, Optional, ClassVar
from numbers import Number

from sint import sint

# %% ../nbs/04_fact.ipynb 4
from .errs import FactorError
from .cons import BASE, ExpoChr
from .util import getmval, dcsrct

# %% ../nbs/04_fact.ipynb 6
class fmeta(type):
    '''Metaclass for SI unit factors.

    `fmeta` acts as a metaclass for SI unit factors. It sets class variables based on
    the superclass and allows each subclass to have its own definitions. It also handles
    the registration of each factor class.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the factor.
    abrv : ClassVar[str]
        The abbreviation of the factor.
    symb : ClassVar[str]
        The symbol of the factor.
    expo : ClassVar[int]
        The exponent of the factor.
    sign : ClassVar[sint]
        The sign of the exponent (Positive, Negative, or Zero).
    ekey : ClassVar[int]
        Calculated as sign * expo.

    Methods
    -------
    clsdef(cls, attr: str, default: Any = None, **kwargs)
        Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first.
    '''  
    def __new__(cls, name, bases, cls_dict):              
        attrs = ('sign', 'expo')
        vdict = dict((attr, getmval(bases, cls_dict, attr)) for attr in attrs)
        sign, expo = dcsrct(vdict, 'sign', 'expo')        
        cls_dict['ekey'] = sign * expo if not any((sign is None, expo is None)) else None
        
        fcls = super().__new__(cls, name, bases, cls_dict)
        return fcls

# %% ../nbs/04_fact.ipynb 8
class fact(float, metaclass=fmeta):
    '''Class for representing SI unit factors.

    `fact` is a class for representing SI unit factors like kilo, mega, giga, etc.
    It is defined by its name, abbreviation, symbol, and exponent. It provides methods
    to convert values between factors, list all existing factors, and more.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the factor.
    abrv : ClassVar[str]
        The abbreviation of the factor.
    symb : ClassVar[str]
        The symbol of the factor.
    expo : ClassVar[int]
        The exponent of the factor.
    sign : ClassVar[sint]
        The sign of the exponent (Positive, Negative, or Zero).
    ekey : ClassVar[int]
        Calculated as sign * expo.

    Instance Attributes
    -------------------
    base : Optional[int]
        The base for the exponent, defaults to 10.
    echr : ExpoChr
        Character used for indicating exponentiation.
    showsign : bool
        Whether to show the sign in the representation.
    showbase : bool
        Whether to show the base in the representation.
    showefmt : bool
        Whether to show the exponent in the representation.

    Methods
    -------
    clsdef(attr: str, default: Any = None, **kwargs)
        Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first.
    
    iseq(v)
        Checks if the factor is equivalent to the value `v`.

    get(v)
        Retrieves the factor corresponding to the value `v`.

    bump(fcls: 'fact', bump: int = 0)
        Bumps the exponent of the factor by `bump`.

    list(attr: str = 'abrv')
        Lists all registered factors based on the attribute `attr`.

    convert(val: Number, to: 'fact')
        Converts a value to a different factor.

    register(f: 'fact')
        Registers a factor class.

    ispos()
        Checks if the factor is positive.
    isneg()
        Checks if the factor is negative.

    Properties
    ----------
    fstr
        Returns the exponential format of the factor, e.g., F3N.

    bstr
        Returns the base format of the factor, e.g., `10^-3`.

    efmt
        Returns the exponential format of the factor, e.g., 10-3.

    ispos()
        Checks if the factor is positive.

    isneg()
        Checks if the factor is negative.
    '''
    # ------------------------------------------------------------------------------------
    # class variables & methods
    # ------------------------------------------------------------------------------------
    name: ClassVar[str]
    abrv: ClassVar[str]
    symb: ClassVar[str]
    expo: ClassVar[int] = 0
    sign: ClassVar[sint] = None
    ekey: ClassVar[int] # NOTE: sign * expo, defined in `fmeta`

    @classmethod
    def clsdef(cls, attr: str, default: Any = None, **kwargs):
        '''Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first'''
        val = kwargs.get(attr, getattr(cls, attr, default))
        if attr == 'echr': val = ExpoChr.to(val)
        if attr == 'ekey': val = cls.sign * cls.expo
        return default if val is None else val
    
    def __new__(cls, base: Optional[int] = BASE, **kwargs) -> 'fact':
        '''Set the float to be the powered value of the base and the exponent'''
        numb = sint.topow(base, cls.expo, cls.sign)
        obj = float.__new__(cls, numb)
        obj.base = base        
        return obj

    def __init__(self, *args, **kwargs) -> None:
        for attr in ('echr', 'showsign', 'showbase', 'showefmt'):
            setattr(self, attr, self.clsdef(attr, **kwargs))

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.register(cls)   

    @classmethod
    def iseq(cls, v) -> bool:
        '''Check if the class or its attributes match a given value.

        Parameters
        ----------
        v : Any
            The value to be checked.

        Returns
        -------
        bool
            True if the value matches the class or any of its attributes (`name`, `abrv`, `symb`, `ekey`), otherwise False.
        '''       
        if cls == v: return True
        keys = ('name', 'abrv', 'symb', 'ekey',)
        if any(v == getattr(cls, attr, None) for attr in keys): return True
        return False

    @classmethod
    def get(cls, v) -> Optional['fact']:
        '''Retrieve a factor class that matches a given value.

        Parameters
        ----------
        v : Any
            The value to be checked.

        Returns
        -------
        Optional['fact']
            Returns the factor class if found, otherwise None.
        '''  
        for f in cls.facts:
            if f.iseq(v): return f
        return None
    
    @classmethod
    def bump(cls, fcls: 'fact', bump: int = 0) -> 'fact':
        '''Bump the exponent of the factor by a given amount.

        Parameters
        ----------
        fcls : 'fact'
            The factor class to bump.
        bump : int, optional
            The amount to bump the exponent by, defaults to 0.

        Returns
        -------
        'fact'
            The new factor class with the bumped exponent.

        Examples
        --------
        >>> tuple((i, fact.moveunit(one, i).name) for i in range(-13, 14, 1))
        '''
        exps = sorted({f.expo for f in cls.facts}) # [0, 1, 2, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30]
        sign, expo = fcls.sign if fcls.sign != sint.Z else 1, fcls.expo
        
        # Get current index of the exponent
        eidx = exps.index(expo)
        
        nidx = ((sign * eidx) + bump)
        if nidx < 0: sign, nidx = -sign, abs(nidx) # Adjust negative indices

        # Handle sign change when crossing boundaries e.g. `<= -30` or `>= +30`
        if len(exps) <= abs(nidx): return fact.get(sign * exps[len(exps) - 1]) 
        return cls.get(sign * exps[nidx]) # Get the new exponent
    
    @classmethod
    def list(cls, attr: str = 'abrv') -> list:
        '''List all registered factors based on a given attribute.

        Parameters
        ----------
        attr : str, optional
            The attribute to list, defaults to 'abrv'.

        Returns
        -------
        list
            A list of all registered factors based on the attribute.
        '''
        return [getattr(f, attr, None) for f in cls.facts]
        
    @classmethod
    def convert(cls, val: Number, to: 'fact') -> float:
        '''Convert a number to a different factor.

        Parameters
        ----------
        val : Number
            The value to convert.
        to : 'fact'
            The factor class to convert to.

        Returns
        -------
        float
            The converted value.

        Raises
        ------
        FactorError
            If the target factor is not found.
        '''
        fcls = cls.get(to)
        if not fcls: raise FactorError(to)
        return val * float(fcls(getattr(val, 'base', cls.base)))
    
    # ------------------------------------------------------------------------------------
    # class / subclass shared variables & methods
    # ------------------------------------------------------------------------------------
    facts: ClassVar[set['fact']] = set()
    @classmethod
    def register(cls, f: 'fact'):
        rattrs = ('name', 'expo', 'sign') # required attributes to be a valid `fact` subclass
        hasall = all(hasattr(f, attr) for attr in rattrs)
        if not hasall and not isinstance(f, cls): return
        cls.facts.add(f)

    # ------------------------------------------------------------------------------------
    # instant variables & methods
    # ------------------------------------------------------------------------------------
    base: Optional[int] = BASE
    echr: ExpoChr  = ExpoChr.carot
    showsign: bool = True
    showbase: bool = False    
    showefmt: bool = True

    @property
    def fstr(self):
        '''Returns the exponential format of the factor e.g. F3N'''
        return sint.make_fstr(self.expo, self.sign, self.base, self.echr, self.showsign, self.showbase)        
    @property
    def bstr(self):
        '''Returns the base format of the factor e.g. `10^-3`'''
        return sint.make_bstr(self.expo, self.sign, self.base, self.echr, self.showsign, self.showbase)        
    @property
    def efmt(self):
        '''Returns the exponential format of the factor e.g. 10-3'''
        return sint.make_estr(self.expo, self.sign, self.base, self.echr, self.showsign, self.showbase)    

    def ispos(self): return sint.to(self.sign) == sint.P
    def isneg(self): return sint.to(self.sign) == sint.N
    def __str__(self) -> str:  return f'{self.efmt}' if self.showefmt else f'{self.fstr}'
    def __repr__(self) -> str: return str(self)

# %% ../nbs/04_fact.ipynb 10
# fact subclasses for positive (`pfact`, e.g. e+<n>), negative (`nfact`, e.g. e-<n>), and the "base" factor (`zfact` e.g. e0)
class zfact(fact): abrv, symb, expo, sign =  None, None, 0, sint.Z
class pfact(fact): sign = sint.P
class nfact(fact): sign = sint.N

# %% ../nbs/04_fact.ipynb 13
# zfact subclass for the "base" factor (e0)
class one(zfact): name = 'one'

# %% ../nbs/04_fact.ipynb 15
# pfact subclasses
class deca(pfact): name, abrv, symb, expo = 'ten', 'deca', 'da', 1
class hecto(pfact): name, abrv, symb, expo = 'hundred', 'hecto', 'h', 2
class kilo(pfact): name, abrv, symb, expo = 'thousand', 'kilo', 'k', 3
class mega(pfact): name, abrv, symb, expo = 'million', 'mega', 'M', 6
class giga(pfact): name, abrv, symb, expo = 'billion', 'giga', 'G', 9
class tera(pfact): name, abrv, symb, expo = 'trillion', 'tera', 'T', 12
class peta(pfact): name, abrv, symb, expo = 'quadrillion', 'peta', 'P', 15
class exa(pfact): name, abrv, symb, expo = 'quintillion', 'exa', 'E', 18
class zetta(pfact): name, abrv, symb, expo = 'sextillion', 'zetta', 'Z', 21
class yotta(pfact): name, abrv, symb, expo = 'septillion', 'yotta', 'Y', 24
class ronna(pfact): name, abrv, symb, expo = 'octillion', 'ronna', 'R', 27
class quetta(pfact): name, abrv, symb, expo = 'nonillion', 'quetta', 'Q', 30

# %% ../nbs/04_fact.ipynb 17
# nfact subclasses
class deci(nfact): name, abrv, symb, expo = 'tenth', 'deci', 'd', 1
class centi(nfact): name, abrv, symb, expo = 'hundredth', 'centi', 'c', 2
class milli(nfact): name, abrv, symb, expo = 'thousandth', 'milli', 'm', 3
class micro(nfact): name, abrv, symb, expo = 'millionth', 'micro', 'u', 6
class nano(nfact): name, abrv, symb, expo = 'billionth', 'nano', 'n', 9
class pico(nfact): name, abrv, symb, expo = 'trillionth', 'pico', 'p', 12
class femto(nfact): name, abrv, symb, expo = 'quadrillionth', 'femto', 'f', 15
class atto(nfact): name, abrv, symb, expo = 'quintillionth', 'atto', 'a', 18
class zepto(nfact): name, abrv, symb, expo = 'sextillionth', 'zepto', 'z', 21
class yocto(nfact): name, abrv, symb, expo = 'septillionth', 'yocto', 'y', 24
class ronto(nfact): name, abrv, symb, expo = 'octillionth', 'ronto', 'r', 27
class quecto(nfact): name, abrv, symb, expo = 'nonillionth', 'quecto', 'q', 30

# %% ../nbs/04_fact.ipynb 19
SI_FACTORS = (    
    one, 
    deca, hecto, kilo, mega, giga, tera, peta, exa, zetta, yotta, ronna, quetta, 
    deci, centi, milli, micro, nano, pico, femto, atto, zepto, yocto, ronto, quecto
)
for f in SI_FACTORS: f.register(f)
fact.facts = set(SI_FACTORS)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_unit.ipynb.

# %% auto 0
__all__ = ['umeta', 'unit']

# %% ../nbs/05_unit.ipynb 3
from typing import Any, Optional, ClassVar
from pstr.fmts import fmtunit

# %% ../nbs/05_unit.ipynb 4
from .atyp import FactRepr
from .errs import FactorError
from .cons import BASE
from .fact import fact, one

# %% ../nbs/05_unit.ipynb 6
class umeta(type):
    '''A metaclass for setting SI unit properties on the unit class.

    The `umeta` metaclass attaches SI unit properties to classes at the time of their creation. 
    It pulls the attributes from the `fact` class and sets them as properties on the class it 
    initializes.

    Attributes
    ----------
    units : set['fact']
        Set of SI unit factors, populated from `fact.facts`.

    Methods
    -------
    __new__(cls, name, bases, cls_dict)
        Create a new instance of the metaclass, setting the unit attributes.

    Examples
    --------
    >>> class unit(float, metaclass=umeta):
    ...     pass
    >>> unit.units
    (one, decka, hecto, ..., quetta, deci, ..., quecto)
    '''
    
    def __new__(cls, name, bases, cls_dict):
        # Generate the attributes for the SI units
        cls_dict['units'] = fact.facts # e.g. (one, decka, hecto, ..., quetta, deci, ..., quecto)
        ucls = super().__new__(cls, name, bases, cls_dict)
        
        for fcls in fact.facts:            
            cfunc = ucls.getcfunc(fcls=fcls)
            for attr in ('abrv', 'name', 'symb'):
                if (aval := getattr(fcls, attr, None)) is not None: 
                    setattr(ucls, aval, property(cfunc))        
        return ucls

# %% ../nbs/05_unit.ipynb 8
class unit(float, metaclass=umeta):
    '''Class for representing SI units, inheriting from float and using `umeta` as the metaclass.

    The `unit` class defines the properties and methods for handling units in the SI system. 
    It allows for conversion between different units and representations.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the unit.
    debug : ClassVar[bool]
        Debugging flag to control representation.

    Instance Attributes
    -------------------
    core : Optional[float]
        The actual float value stored.
    base : Optional[int]
        The base for the exponent, defaults to 10.
    fcls : Optional[fact]
        The factor class being used, defaults to `one`.
    numb : Optional[float]
        The number used for conversion.
    ndig : Optional[int]
        Number of digits to display.
    shownumb : Optional[bool]
        Whether to show the number in representation.
    abrvunit : Optional[bool]
        Whether to abbreviate the unit.
    factrepr : Optional[FactRepr]
        How to represent the factor ('abrv', 'name', 'symb').

    Methods
    -------
    getcfunc(fcls: fact)
        Returns a class method for converting the unit to the given factor class.
    factor()
        Actually applies the factor to the base.
    to(key, inplace: bool = False)
        Converts the unit to a different factor.
    '''
    # ------------------------------------------------------------------------------------
    # class variables & methods
    # ------------------------------------------------------------------------------------
    name: ClassVar[str]
    debug: ClassVar[bool] = False
    @classmethod
    def getcfunc(cls: type, fcls: fact):
        '''returns a class method that converts the unit to the given factor class'''
        def cfunc(self, inplace: bool = False):
            return self.to(fcls, inplace)
        return cfunc

    # ------------------------------------------------------------------------------------
    # instant variables & methods
    # ------------------------------------------------------------------------------------
    core: Optional[float]        # NOTE: this is the number that is actually stored in the float    
    base: Optional[int] = BASE   # base to use for the conversion
    fcls: Optional[fact] = one   # factor class to use, defaults to `one` (e.g. e0=1) so that `unit` works as float as expected by default
    numb: Optional[float] = None # NOTE: this is the number used for the conversion
    ndig: Optional[int] = 3      # number of digits to show

    shownumb: Optional[bool] = True
    abrvunit: Optional[bool] = True
    factrepr: Optional[FactRepr] = 'symb'

    def __new__(cls, core: float, *args, **kwargs):
        obj = float.__new__(cls, core)
        return obj
    
    @classmethod
    def clsdef(cls, attr: str, default: Any = None, **kwargs):
        '''Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first'''
        val = kwargs.get(attr, getattr(cls, attr, default))
        return default if val is None else val

    def __init__(self, core: float, **kwargs):
        attrs = ('numb', 'base', 'fcls', 'ndig', 'shownumb', 'abrvunit', 'factrepr')
        adefs = tuple(self.clsdef(attr, (core if attr == 'numb' else None), **kwargs) for attr in attrs)
        for i, attr in enumerate(attrs):
            setattr(self, attr, adefs[i])


    @property
    def factor(self) -> fact:
        '''actually applies the factor to the base'''
        return self.fcls(self.base)

    def __str__(self) -> str:
        ustr = fmtunit(
            org = float(self.numb), flt = float(self), unt = self, fct = self.fcls,
            factrepr = self.factrepr, shownumb = self.shownumb,
            abrvunit = self.abrvunit, unitname = self.name, ndig = self.ndig
        )
        return ustr
    
    def __repr__(self) -> str:
        if type(self).debug:
            return f'{float(self)} (o={self.numb}, b={self.base}, f={self.fcls.__name__})'        
        return str(self)

    def to(self, key, inplace: bool = False):
        '''Converts the unit to a different factor.

        This method changes the unit of the number to the unit specified by the `key`.
        The conversion is either done in-place or a new object is returned based on the
        `inplace` parameter.

        Parameters
        ----------
        key : Any
            The key used to look up the factor class. It could be an abbreviation, name, 
            symbol, or the factor class itself.
        inplace : bool, optional
            If True, the conversion is done in-place and the object itself is modified.
            Otherwise, a new object with the converted value is returned. Default is False.

        Returns
        -------
        unit
            A new `unit` object with the converted value if `inplace` is False, otherwise 
            modifies the object itself and returns it.

        Raises
        ------
        FactorError
            If the provided `key` is not found in the available factors.

        Examples
        --------
        >>> u = unit(1000, base=10, fcls=kilo)
        >>> u.to(mega)
        unit(1)
        >>> u.to(mega, inplace=True)
        unit(1)
        '''
        fcls = fact.get(key)
        if not fcls: raise FactorError(key)
        fcls = fcls(self.base)
        curr = self.numb / fcls
        if inplace:
            return type(self)(core=curr, numb=curr, base=self.base, fcls=one(self.base))        
        return type(self)(core=curr, numb=self.numb, base=self.base, fcls=fcls)
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_atyp.ipynb.

# %% auto 0
__all__ = ['FactQ', 'FactRepr']

# %% ../nbs/00_atyp.ipynb 3
from typing import Literal, Optional, TypeAlias

# %% ../nbs/00_atyp.ipynb 5
FactQ: TypeAlias = Optional['fact']
FactRepr: TypeAlias = Literal['abrv', 'name', 'symb']



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_docs.ipynb.

# %% auto 0
__all__ = ['fmeta', 'fact', 'iseq', 'get', 'bump', 'flist', 'convert', 'umeta', 'unit', 'to', 'fmtunit']

# %% ../nbs/04_docs.ipynb 3
#| export

# %% ../nbs/04_docs.ipynb 6
fmeta = '''Metaclass for SI unit factors.

    `fmeta` acts as a metaclass for SI unit factors. It sets class variables based on
    the superclass and allows each subclass to have its own definitions. It also handles
    the registration of each factor class.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the factor.
    abrv : ClassVar[str]
        The abbreviation of the factor.
    symb : ClassVar[str]
        The symbol of the factor.
    expo : ClassVar[int]
        The exponent of the factor.
    sign : ClassVar[sint]
        The sign of the exponent (Positive, Negative, or Zero).
    ekey : ClassVar[int]
        Calculated as sign * expo.

    Methods
    -------
    clsdef(cls, attr: str, default: Any = None, **kwargs)
        Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first.
    '''
fact = '''Class for representing SI unit factors.

    `fact` is a class for representing SI unit factors like kilo, mega, giga, etc.
    It is defined by its name, abbreviation, symbol, and exponent. It provides methods
    to convert values between factors, list all existing factors, and more.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the factor.
    abrv : ClassVar[str]
        The abbreviation of the factor.
    symb : ClassVar[str]
        The symbol of the factor.
    expo : ClassVar[int]
        The exponent of the factor.
    sign : ClassVar[sint]
        The sign of the exponent (Positive, Negative, or Zero).
    ekey : ClassVar[int]
        Calculated as sign * expo.

    Instance Attributes
    -------------------
    base : Optional[int]
        The base for the exponent, defaults to 10.
    echr : ExpoChr
        Character used for indicating exponentiation.
    showsign : bool
        Whether to show the sign in the representation.
    showbase : bool
        Whether to show the base in the representation.
    showefmt : bool
        Whether to show the exponent in the representation.

    Methods
    -------
    clsdef(attr: str, default: Any = None, **kwargs)
        Get the `attr` from `kwargs`, the class, or the `default`, whichever is found first.
    
    iseq(v)
        Checks if the factor is equivalent to the value `v`.

    get(v)
        Retrieves the factor corresponding to the value `v`.

    bump(fcls: 'fact', bump: int = 0)
        Bumps the exponent of the factor by `bump`.

    list(attr: str = 'abrv')
        Lists all registered factors based on the attribute `attr`.

    convert(val: Number, to: 'fact')
        Converts a value to a different factor.

    register(f: 'fact')
        Registers a factor class.

    ispos()
        Checks if the factor is positive.
    isneg()
        Checks if the factor is negative.

    Properties
    ----------
    fstr
        Returns the exponential format of the factor, e.g., F3N.

    bstr
        Returns the base format of the factor, e.g., `10^-3`.

    efmt
        Returns the exponential format of the factor, e.g., 10-3.

    ispos()
        Checks if the factor is positive.

    isneg()
        Checks if the factor is negative.
    '''
iseq = '''Check if the class or its attributes match a given value.

        Parameters
        ----------
        v : Any
            The value to be checked.

        Returns
        -------
        bool
            True if the value matches the class or any of its attributes (`name`, `abrv`, `symb`, `ekey`), otherwise False.
        '''
get = '''Retrieve a factor class that matches a given value.

        Parameters
        ----------
        v : Any
            The value to be checked.

        Returns
        -------
        Optional['fact']
            Returns the factor class if found, otherwise None.
        '''    
bump = '''Bump the exponent of the factor by a given amount.

        Parameters
        ----------
        fcls : 'fact'
            The factor class to bump.
        bump : int, optional
            The amount to bump the exponent by, defaults to 0.

        Returns
        -------
        'fact'
            The new factor class with the bumped exponent.

        Examples
        --------
        >>> tuple((i, fact.moveunit(one, i).name) for i in range(-13, 14, 1))
        '''
flist = '''List all registered factors based on a given attribute.

        Parameters
        ----------
        attr : str, optional
            The attribute to list, defaults to 'abrv'.

        Returns
        -------
        list
            A list of all registered factors based on the attribute.
        '''
convert = '''Convert a number to a different factor.

        Parameters
        ----------
        val : Number
            The value to convert.
        to : 'fact'
            The factor class to convert to.

        Returns
        -------
        float
            The converted value.

        Raises
        ------
        FactorError
            If the target factor is not found.
        '''

# %% ../nbs/04_docs.ipynb 8
umeta = '''A metaclass for setting SI unit properties on the unit class.

    The `umeta` metaclass attaches SI unit properties to classes at the time of their creation. 
    It pulls the attributes from the `fact` class and sets them as properties on the class it 
    initializes.

    Attributes
    ----------
    units : set['fact']
        Set of SI unit factors, populated from `fact.facts`.

    Methods
    -------
    __new__(cls, name, bases, cls_dict)
        Create a new instance of the metaclass, setting the unit attributes.

    Examples
    --------
    >>> class unit(float, metaclass=umeta):
    ...     pass
    >>> unit.units
    (one, decka, hecto, ..., quetta, deci, ..., quecto)
    '''

unit = '''Class for representing SI units, inheriting from float and using `umeta` as the metaclass.

    The `unit` class defines the properties and methods for handling units in the SI system. 
    It allows for conversion between different units and representations.

    Class Attributes
    ----------------
    name : ClassVar[str]
        The name of the unit.
    debug : ClassVar[bool]
        Debugging flag to control representation.

    Instance Attributes
    -------------------
    core : Optional[float]
        The actual float value stored.
    base : Optional[int]
        The base for the exponent, defaults to 10.
    fcls : Optional[fact]
        The factor class being used, defaults to `one`.
    numb : Optional[float]
        The number used for conversion.
    ndig : Optional[int]
        Number of digits to display.
    shownumb : Optional[bool]
        Whether to show the number in representation.
    abrvunit : Optional[bool]
        Whether to abbreviate the unit.
    factrepr : Optional[FactRepr]
        How to represent the factor ('abrv', 'name', 'symb').

    Methods
    -------
    getcfunc(fcls: fact)
        Returns a class method for converting the unit to the given factor class.
    factor()
        Actually applies the factor to the base.
    to(key, inplace: bool = False)
        Converts the unit to a different factor.
    '''


to = '''Converts the unit to a different factor.

        This method changes the unit of the number to the unit specified by the `key`.
        The conversion is either done in-place or a new object is returned based on the
        `inplace` parameter.

        Parameters
        ----------
        key : Any
            The key used to look up the factor class. It could be an abbreviation, name, 
            symbol, or the factor class itself.
        inplace : bool, optional
            If True, the conversion is done in-place and the object itself is modified.
            Otherwise, a new object with the converted value is returned. Default is False.

        Returns
        -------
        unit
            A new `unit` object with the converted value if `inplace` is False, otherwise 
            modifies the object itself and returns it.

        Raises
        ------
        FactorError
            If the provided `key` is not found in the available factors.

        Examples
        --------
        >>> u = unit(1000, base=10, fcls=kilo)
        >>> u.to(mega)
        unit(1)
        >>> u.to(mega, inplace=True)
        unit(1)
        '''

fmtunit = '''Formats a given number to its representation in units with factors.

    Parameters
    ----------
    org : float
        The number as converted to the desired factor.
    flt : Optional[float], default=None
        The original number prior to factor conversion.
    unt : Optional[unit], default=None
        An instance of a unit class representing the base unit.
    fct : Optional[fact], default=None
        The factor class for the desired unit representation.
    shownumb : Optional[bool], default=True
        Whether to show the original number (`flt`) or the scaled number (`org`).
    abrvunit : Optional[bool], default=False
        Whether to show the abbreviation or the full name of the unit.
    factrepr : Optional[str], default='abrv'
        Representation of factor class to use, options include 'name', 'abrv', and 'symb'.
    ndig : Optional[int], default=3
        Number of significant digits to represent the number.
    unitname : Optional[str], default=None
        Override for the unit's name.
    factname : Optional[str], default=None
        Override for the factor's name or representation.

    Returns
    -------
    str
        Formatted string representation of the number in units with factors.

    Examples
    --------
    >>> fmtunit(1000.0, flt=1, unt=second, fct=kilo, factrepr='name')
    '1 Kilosecond'
    >>> fmtunit(1000.0, flt=1, unt=second, fct=kilo, factrepr='abrv')
    '1 Ksecond'
    >>> fmtunit(1000.0, flt=1, unt=second, fct=kilo, factrepr='symb')
    '1 KS'
    '''



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_cons.ipynb.

# %% auto 0
__all__ = ['BASE', 'SIGNS', 'EXPOS', 'ExpoChr']

# %% ../nbs/01_cons.ipynb 3
from enum import StrEnum

from sint import sint
from pstr.nchr import CARET, STARSTAR

# %% ../nbs/01_cons.ipynb 5
BASE = 10
SIGNS = (sint.P, sint.N)
EXPOS = (1, 2, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30)

class ExpoChr(StrEnum):
    carot = CARET    # '^'
    dstar = STARSTAR # '**'
    @classmethod
    def to(cls, v: str): return cls(v) if cls.isin(v) else cls.carot
    @classmethod
    def isin(cls, v: str): return v in list(map(str, cls._member_map_.values()))



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/nist',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/nist',
                'lib_path': 'nist'},
  'syms': { 'nist.atyp': {},
            'nist.cons': { 'nist.cons.ExpoChr': ('cons.html#expochr', 'nist/cons.py'),
                           'nist.cons.ExpoChr.isin': ('cons.html#expochr.isin', 'nist/cons.py'),
                           'nist.cons.ExpoChr.to': ('cons.html#expochr.to', 'nist/cons.py')},
            'nist.docs': {},
            'nist.errs': { 'nist.errs.FactorError': ('errs.html#factorerror', 'nist/errs.py'),
                           'nist.errs.FactorError.__init__': ('errs.html#factorerror.__init__', 'nist/errs.py')},
            'nist.fact': { 'nist.fact.atto': ('fact.html#atto', 'nist/fact.py'),
                           'nist.fact.centi': ('fact.html#centi', 'nist/fact.py'),
                           'nist.fact.deca': ('fact.html#deca', 'nist/fact.py'),
                           'nist.fact.deci': ('fact.html#deci', 'nist/fact.py'),
                           'nist.fact.exa': ('fact.html#exa', 'nist/fact.py'),
                           'nist.fact.fact': ('fact.html#fact', 'nist/fact.py'),
                           'nist.fact.fact.__init__': ('fact.html#fact.__init__', 'nist/fact.py'),
                           'nist.fact.fact.__init_subclass__': ('fact.html#fact.__init_subclass__', 'nist/fact.py'),
                           'nist.fact.fact.__new__': ('fact.html#fact.__new__', 'nist/fact.py'),
                           'nist.fact.fact.__repr__': ('fact.html#fact.__repr__', 'nist/fact.py'),
                           'nist.fact.fact.__str__': ('fact.html#fact.__str__', 'nist/fact.py'),
                           'nist.fact.fact.bstr': ('fact.html#fact.bstr', 'nist/fact.py'),
                           'nist.fact.fact.bump': ('fact.html#fact.bump', 'nist/fact.py'),
                           'nist.fact.fact.clsdef': ('fact.html#fact.clsdef', 'nist/fact.py'),
                           'nist.fact.fact.convert': ('fact.html#fact.convert', 'nist/fact.py'),
                           'nist.fact.fact.efmt': ('fact.html#fact.efmt', 'nist/fact.py'),
                           'nist.fact.fact.fstr': ('fact.html#fact.fstr', 'nist/fact.py'),
                           'nist.fact.fact.get': ('fact.html#fact.get', 'nist/fact.py'),
                           'nist.fact.fact.iseq': ('fact.html#fact.iseq', 'nist/fact.py'),
                           'nist.fact.fact.isneg': ('fact.html#fact.isneg', 'nist/fact.py'),
                           'nist.fact.fact.ispos': ('fact.html#fact.ispos', 'nist/fact.py'),
                           'nist.fact.fact.list': ('fact.html#fact.list', 'nist/fact.py'),
                           'nist.fact.fact.register': ('fact.html#fact.register', 'nist/fact.py'),
                           'nist.fact.femto': ('fact.html#femto', 'nist/fact.py'),
                           'nist.fact.fmeta': ('fact.html#fmeta', 'nist/fact.py'),
                           'nist.fact.fmeta.__new__': ('fact.html#fmeta.__new__', 'nist/fact.py'),
                           'nist.fact.giga': ('fact.html#giga', 'nist/fact.py'),
                           'nist.fact.hecto': ('fact.html#hecto', 'nist/fact.py'),
                           'nist.fact.kilo': ('fact.html#kilo', 'nist/fact.py'),
                           'nist.fact.mega': ('fact.html#mega', 'nist/fact.py'),
                           'nist.fact.micro': ('fact.html#micro', 'nist/fact.py'),
                           'nist.fact.milli': ('fact.html#milli', 'nist/fact.py'),
                           'nist.fact.nano': ('fact.html#nano', 'nist/fact.py'),
                           'nist.fact.nfact': ('fact.html#nfact', 'nist/fact.py'),
                           'nist.fact.one': ('fact.html#one', 'nist/fact.py'),
                           'nist.fact.peta': ('fact.html#peta', 'nist/fact.py'),
                           'nist.fact.pfact': ('fact.html#pfact', 'nist/fact.py'),
                           'nist.fact.pico': ('fact.html#pico', 'nist/fact.py'),
                           'nist.fact.quecto': ('fact.html#quecto', 'nist/fact.py'),
                           'nist.fact.quetta': ('fact.html#quetta', 'nist/fact.py'),
                           'nist.fact.ronna': ('fact.html#ronna', 'nist/fact.py'),
                           'nist.fact.ronto': ('fact.html#ronto', 'nist/fact.py'),
                           'nist.fact.tera': ('fact.html#tera', 'nist/fact.py'),
                           'nist.fact.yocto': ('fact.html#yocto', 'nist/fact.py'),
                           'nist.fact.yotta': ('fact.html#yotta', 'nist/fact.py'),
                           'nist.fact.zepto': ('fact.html#zepto', 'nist/fact.py'),
                           'nist.fact.zetta': ('fact.html#zetta', 'nist/fact.py'),
                           'nist.fact.zfact': ('fact.html#zfact', 'nist/fact.py')},
            'nist.unit': { 'nist.unit.umeta': ('unit.html#umeta', 'nist/unit.py'),
                           'nist.unit.umeta.__new__': ('unit.html#umeta.__new__', 'nist/unit.py'),
                           'nist.unit.unit': ('unit.html#unit', 'nist/unit.py'),
                           'nist.unit.unit.__init__': ('unit.html#unit.__init__', 'nist/unit.py'),
                           'nist.unit.unit.__new__': ('unit.html#unit.__new__', 'nist/unit.py'),
                           'nist.unit.unit.__repr__': ('unit.html#unit.__repr__', 'nist/unit.py'),
                           'nist.unit.unit.__str__': ('unit.html#unit.__str__', 'nist/unit.py'),
                           'nist.unit.unit.clsdef': ('unit.html#unit.clsdef', 'nist/unit.py'),
                           'nist.unit.unit.factor': ('unit.html#unit.factor', 'nist/unit.py'),
                           'nist.unit.unit.getcfunc': ('unit.html#unit.getcfunc', 'nist/unit.py'),
                           'nist.unit.unit.to': ('unit.html#unit.to', 'nist/unit.py')},
            'nist.util': { 'nist.util.dcsrct': ('util.html#dcsrct', 'nist/util.py'),
                           'nist.util.getbval': ('util.html#getbval', 'nist/util.py'),
                           'nist.util.getmval': ('util.html#getmval', 'nist/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_errs.ipynb.

# %% auto 0
__all__ = ['FactorError']

# %% ../nbs/02_errs.ipynb 3
#| export

# %% ../nbs/02_errs.ipynb 5
class FactorError(KeyError):
    def __init__(self, key: str):  
        super().__init__(f'No SI unit associated with key: {key}')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_util.ipynb.

# %% auto 0
__all__ = ['getbval', 'getmval', 'dcsrct']

# %% ../nbs/03_util.ipynb 3
from typing import Any, Optional

# %% ../nbs/03_util.ipynb 5
def getbval(bases: tuple, attr: str, default = None) -> Optional[Any]:
    '''Get the first attribute value found in the base classes'''
    for base in bases:
        if hasattr(base, attr): return getattr(base, attr)
    return default

def getmval(bases: tuple, cls_dict: dict, attr: str, default = None) -> Optional[Any]:
    '''Get the first attribute either from the `cls_dict` or in any of the base classes'''
    cval = cls_dict.get(attr)
    bval = getbval(bases, attr, default)
    return cval if cval is not None else bval

def dcsrct(d: dict, *args, __flbk: Any = None) -> tuple:
    '''Deconstruct a dictionary
    
    Examples
    --------
    >>> b, d = dcsrct(dict(a=1, b=2, c=3, d=4), 'b', 'd')
    >>> b, d
    (2, 4)
    '''
    if __flbk is None: __flbk = tuple(None for _ in range(len(args)))
    return tuple(d.get(attr, dflt) for attr, dflt in zip(args, __flbk))
